/**
  Tiny OpenGL/GLUT demo of a reaction-diffusion system: water freezing into ice.
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2009-02-17 (Public Domain)
*/
#include "GL/glew.h" /* GL extension wrangler: http://glew.sourceforge.net/ */
#include "GL/glut.h" /* windows, mice, etc.  http://freeglut.sourceforge.net */
#include <stdlib.h> /* for "rand" */
#include <stdio.h> /* for printf */
#include "osl/vec4.h" /* 3D vector class */
#include "GL/glui.h" /* user interface widgets: http://glui.sourceforge.net/ */

#include "soil/SOIL.h" /* Simple OpenGL Image Library, www.lonesock.net/soil.html (plus Dr. Lawlor's modifications) */
#include "soil/SOIL.c" /* just slap in implementation files here, for easier linking */
#include "soil/stb_image_aug.c"

#include "ogl/glsl.h"
#include "ogl/glsl.cpp"

#define VEC3_TO_XYZ(v) (v).x,(v).y,(v).z /* convert a vec3 to three float arguments */
#include <vector> /* std::vector container */

#ifndef M_PI /* why doesn't windows math.h have this? */
#define M_PI 3.14159265358979
#endif


/* These variables control the texture filtering happening below */
/* Enable for anisotropic filtering */
int texaniso=1;
/* Texture pixel format on graphics card */
int texformat=GL_RGBA8; 
/* Texture "wrap mode" in X and Y: */
int wrap_mode=GL_REPEAT;
/* Texture magnification mode/minification mode */
int minfilter_mode=GL_LINEAR_MIPMAP_LINEAR; 
int magfilter_mode=GL_LINEAR; 


/* Random Float: Return a random number between 0 and 1 */
float rf(void) { return (rand()&0xfff)*(1.0/0xfff); }
/* Random Vector: return a random vector in 3D, all axes between 0 and 1 */
vec3 rv(void) {return vec3(rf(),rf(),rf());}

bool key_down[256]; /* if true, the corresponding key is down */

/* Return the wall clock time, in seconds */
double oglTimer(void) {return 0.001*glutGet(GLUT_ELAPSED_TIME);}
/* Check if an error has happened--if so, exit */
void oglCheck(const char *where) {
	int err=0;
	if (0!=(err=glGetError())) {
		fprintf(stderr,"OpenGL error %d at %s\n",err,where);
		exit(1);
	}
}

/* 3D orientation */
/* Orthonormal coordinate frame */
class ortho_frame {
public:
	/* Unit vectors pointing along axes of our frame.
		X cross Y is Z. (right-handed coordinate system)
	*/
	vec3 x,y,z;
	ortho_frame() :x(1,0,0), y(0,1,0), z(0,0,1) {} /* point down Y axis */
	
	/* Reorient this coordinate frame by this far along in the X and Y axes.  
	   "dx" is the distance along the z axis to push the x axis; 
	   "dy" the distance along the z axis to push the y axis.
	*/
	void nudge(double dx,double dy) {
		x+=dx*z;
		y+=dy*z;
		orthonormalize();
	}
	
	/* Reconstruct an orthonormal frame from X and Y axes.
	  Y is primary, X is secondary, Z is tertiary.
	*/
	void orthonormalize(void) {
		y=normalize(y);
		z=normalize(cross(x,y));
		x=normalize(cross(y,z));
	}
};
ortho_frame orient; /* camera orientation */
vec3 camera(0.5,0.5,0.75); /* (initial) location of camera */



/* Make an OpenGL texture handle, of this size, with this pixel format */
GLuint make_texture(int w,int h,GLenum format=GL_RGBA8)
{
	GLuint tex=0;
	glGenTextures(1,&tex);
	GLenum t=GL_TEXTURE_2D;
	glBindTexture(t,tex);
	char *mem=new char[w*h*4]; memset(mem,0x80,w*h*4);
	glTexImage2D(t,0,format, w,h,0, GL_RGBA,GL_UNSIGNED_BYTE,mem);
	delete[] mem;
	glTexParameteri(t, GL_TEXTURE_MAG_FILTER,GL_LINEAR); 
	glTexParameteri(t, GL_TEXTURE_MIN_FILTER,GL_LINEAR); /* no mipmaps */
	GLenum texWrap=GL_CLAMP_TO_EDGE; /* stop changing at edge */
	glTexParameteri(t, GL_TEXTURE_WRAP_S, texWrap); 
	glTexParameteri(t, GL_TEXTURE_WRAP_T, texWrap);
	glBindTexture(t,0); /* back to default texture */
	oglCheck("after make_texture");
	return tex;
}

/* Make a framebuffer object that renders pixels into this texture */
GLuint make_fbo(GLuint tex)
{
	if (glGenFramebuffersEXT==0) {
		printf("You don't have framebuffers!  Goodbye.\n");
		exit(1);
	}
	GLuint fb=0;
	glGenFramebuffersEXT(1, &fb);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, 
		GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, tex, 0);
	
	/* Make sure the framebuffer's happy */
	GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	if (status!=GL_FRAMEBUFFER_COMPLETE_EXT)
	{ fprintf(stderr,"Error %d in framebuffer creation!\n",status); exit(1); }

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0); /* back to screen render */
	oglCheck("after make_fbo");
	return fb;
}

void display(void) /* Draw everything in our window */
{
	oglCheck("before display function");
	static double elapsed=0.0; /* last elapsed display time (seconds) */
	static float dt=0.01, display_dt=0.0;
	static float sim_time=0.0;
	float big=300; /* size of the ground */
	float floor=0.0; /* Z height of floor */
	glFinish(); double start_t=oglTimer();
	
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	
/* Do offscreen rendering work */
	/* load an image file directly as a new OpenGL texture */
	static GLuint srcTex = SOIL_load_OGL_texture
	(
		"image.jpg",
		SOIL_LOAD_AUTO,
		SOIL_CREATE_NEW_ID,
		SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y
	);
	int w=1024, h=1024; /* MUST match image size above (silly!) */
	static GLuint destTex = make_texture(w,h,GL_RGBA8);
	static GLuint fb = make_fbo(destTex);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, 
		GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, destTex, 0);
	
	glViewport(0,0,w,h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	/* From 0..1 texture coords to -1..+1 clip coords */
	glTranslatef(-1.0,-1.0,-1.0);
	glScalef(2.0,2.0,2.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glBindTexture(GL_TEXTURE_2D,srcTex);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);

	/* Actual physics happens right on the graphics card */
	static GLhandleARB prog = makeProgramObject(
	"varying vec2 texCoords;\n"
	"void main(void) /* vertex shader */ {\n"
	"	texCoords = vec2(gl_MultiTexCoord0);\n"
	"	gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;\n"
	"}\n"
	,
	"varying vec2 texCoords;\n"
	"uniform sampler2D srcTex;\n"
	"const int dirs=3; /* MUST match C++ copy below! */\n"
	"uniform vec2 crystalDirs[dirs]; /* texture coordinate shifts in each direction */\n"
	"void main(void) /* fragment shader */ {\n"
	"	float outtemp = texture2DLod(srcTex,texCoords,2.5).r; /* mip-blurred heat */\n"
	"	outtemp -= 0.5/255.0; /* apply gentle cooling */\n"
	
	"	vec4 me=texture2D(srcTex,texCoords);\n"
	"	float nborice=0.0, nbordir=me.g;\n"
	"	for (int d=0;d<dirs;d++) {\n"
	"		vec4 nbor=texture2D(srcTex,texCoords+crystalDirs[d]);\n"
	"		nborice=max(nborice,nbor.b);\n"
	"		if (nbor.b>0.5) nbordir=d*(1.0/float(dirs)); /* for freeze direction */\n"
	"	}\n"
	"	if (me.r < 0.3 && nborice>0.3 && me.b<1.0) \n"
	"	{ /* I'm cold, I'm near ice, but I'm not ice: I can freeze a little */ \n"
	"		me.b += 0.1; /* my ice concentration goes up */\n"
	"		outtemp += 1.0; /* I release the heat of fusion */\n"
	"		me.g = nbordir; /* record which direction I last froze from */ \n"
	"	}\n"
	"	gl_FragColor = vec4(outtemp,me.g,me.b,1.0);\n"
	"}\n"
	);
	glUseProgramObjectARB(prog);
	glUniform1iARB(glGetUniformLocationARB(prog,"srcTex"),0); /* texture unit 0 */
	enum {dirs=3}; /* MUST match GLSL version above! */
	float crystalDirs[2*dirs];
	for (int d=0;d<dirs;d++) {
		float r=1.0/w; /* filter size, in texture coordinates */
		float angle=1.4+d*(2.0*M_PI/dirs); /* filter orientation, radians */
		crystalDirs[2*d+0]=r*cos(angle);
		crystalDirs[2*d+1]=r*sin(angle);
		//printf("crystalDirs[%d] = %f, %f\n",d,crystalDirs[2*d+0],crystalDirs[2*d+1]);
	}
	glUniform2fvARB(glGetUniformLocationARB(prog,"crystalDirs"),dirs,crystalDirs);
	
	/* run this program over the whole rectangle of the texture */
	glBegin(GL_QUADS);
	glColor4f(1,1,1,1); 
	glTexCoord2f(0,0); glVertex2fv(vec3(0,0,0));
	glTexCoord2f(1,0); glVertex2fv(vec3(1,0,0));
	glTexCoord2f(1,1); glVertex2fv(vec3(1,1,0));
	glTexCoord2f(0,1); glVertex2fv(vec3(0,1,0));
	glEnd();
	
	glUseProgramObjectARB(0);

#ifdef __unix__
	usleep(10*1000); /* sleep for 10ms, to cut framerate to 100hz */
#else /* windows? */
	Sleep(10);
#endif	

	glDisable(GL_TEXTURE_2D);
	
	std::swap(srcTex,destTex); /* ping-pong the buffers */
	
	oglCheck("After offscreen work");
	
/* Main camera setup */
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	glViewport(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
	glClearColor(0.3,0.5,0.7,1.0); /* clear to a nice blue color */
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(70.0, /* degree FOV */
		glutGet(GLUT_WINDOW_WIDTH)/float(glutGet(GLUT_WINDOW_HEIGHT)),
		0.1,1000.0 /* Z range */
	);
	
	/* Physics and display timesteps: */
	float vel=0.5*display_dt; /* meters camera motion per frame */
	if (key_down[(int)'w']) camera-=vel*orient.z;
	if (key_down[(int)'s']) camera+=vel*orient.z;
	if (key_down[(int)'a']) camera-=vel*orient.x;
	if (key_down[(int)'d']) camera+=vel*orient.x;
	
	/* Don't allow head tilting! */
	orient.x.z=0.0; 
	orient.y=cross(orient.z,orient.x);
	orient.orthonormalize();
	
	vec3 target=camera-orient.z; /* camera looks down *negative* Z */
	gluLookAt(VEC3_TO_XYZ(camera), 
		VEC3_TO_XYZ(target), 
		VEC3_TO_XYZ(orient.y)
	);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
// Draw in our ground plane texture
	GLenum t=GL_TEXTURE_2D;
	glBindTexture(t,srcTex);
	glGenerateMipmapEXT(t); /* make mipmaps before rendering */
	glEnable(t);
	
	glTexParameteri(t, GL_TEXTURE_MAG_FILTER,magfilter_mode); 
	glTexParameteri(t, GL_TEXTURE_MIN_FILTER,minfilter_mode);
	glTexParameteri(t, GL_TEXTURE_WRAP_S, wrap_mode); 
	glTexParameteri(t, GL_TEXTURE_WRAP_T, wrap_mode);
	glTexParameteri(t, GL_TEXTURE_MAX_ANISOTROPY_EXT,texaniso?16:1);
	
	glDisable(GL_BLEND);
	glBegin(GL_QUADS);
	glColor4f(1,1,1,1); 
	glTexCoord2f(-big,-big); glVertex3f(-big,-big,floor);
	glTexCoord2f(+big,-big); glVertex3f(+big,-big,floor);
	glTexCoord2f(+big,+big); glVertex3f(+big,+big,floor);
	glTexCoord2f(-big,+big); glVertex3f(-big,+big,floor);
	glEnd();
	glDisable(GL_TEXTURE_2D);
	
/* Draw axes */
	for (int axis=0;axis<3;axis++) {
		vec3 color(0.0);
		color[axis]=1.0;
		glColor3fv(color);
		
		glBegin(GL_LINES);
		glVertex3f(0,0,0);
		vec3 end(0.0); end[axis]=1.0;
		glVertex3fv(end);
		glEnd();
		
		glRasterPos3fv(end);
		glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_24,'X'+axis);
	}
	
	glutSwapBuffers(); /* display everything we've rendered so far */
	glutPostRedisplay(); /* continual animation */
	
	// Print frames per second
	glFinish(); elapsed=oglTimer()-start_t; /* stop the stopwatch */
	static double last_print=0.0; /* Periodically print out the time */
	if (start_t>last_print+0.1) {
		display_dt=elapsed+0.001;
		printf("%.1f frames/sec, %.2f ms/frame\n",
			1.0/elapsed,1000.0*elapsed);
		last_print=start_t;
	}
	oglCheck("after display function");
}

int mouse_x=0,mouse_y=0;
void mouse(int button,int state,int x,int y) 
{ /* mouse being pressed or released--save position for motion */
	mouse_x=x; mouse_y=y;
}
void motion(int x, int y) { 
	float scale=0.01; /* radians rotation/frame per mouse pixel motion */
	orient.nudge(
		-scale*(x-mouse_x), 
		scale*(y-mouse_y)
	);
	mouse_x=x; mouse_y=y; /* save old mouse positions */
}
void keyboard(unsigned char key,int x,int y) {
	key_down[key]=true;
}
void keyboardUp(unsigned char key,int x,int y) {
	key_down[key]=false;
}

int main(int argc,char **argv) {
	glutInit(&argc,argv);  /* set up GLUT */
	glutInitWindowPosition(0,60);
	glutInitWindowSize(1000,700);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	int win=glutCreateWindow("Crystal Growth Demo"); /* make a window */
	glewInit(); /* only AFTER create window! */
	
	glutDisplayFunc(display); /* here's how to draw the window */
	glutMouseFunc(mouse);
	glutMotionFunc(motion); 
	glutKeyboardFunc(keyboard);
	glutKeyboardUpFunc(keyboardUp);
	glutMainLoop(); /* start drawing */
}
